#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda_runtime.h>
#include <vector>
#include <string>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

// Warp-level reduction using __shfl_down_sync; avoids using shared memory for small reductions.
__inline__ __device__ float warpReduceSum(float val) {
    // Full mask for active threads within the warp
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        val += __shfl_down_sync(0xffffffff, val, offset);
    }
    return val;
}

// Custom CUDA kernel to perform global average pooling (for output size 1) using warp-level primitives.
// Each block computes one output element corresponding to one (n, c) pair; the kernel processes H*W elements.
__global__ void globalAvgPoolKernel(const float* __restrict__ input,
                                     float* __restrict__ output,
                                     int C,
                                     int HW) {
    // Each block corresponds to one (n, c) pair. Grid dim.x = N * C.
    int idx = blockIdx.x;  // global index for (n, c)
    int lane = threadIdx.x; // thread index within the warp

    // Decode sample index and channel from block index
    int n = idx / C;
    int c = idx % C;

    // Compute starting offset for the (n, c) feature map
    int offset = (n * C + c) * HW;
    float sum = 0.0f;

    // Each thread processes a subset of the H*W elements with stride equal to blockDim.x
    for (int i = lane; i < HW; i += blockDim.x) {
        sum += input[offset + i];
    }

    // Perform warp-level reduction across the threads in the block
    sum = warpReduceSum(sum);

    // Use atomic operation to accumulate results from different warps
    if (lane == 0) {
        atomicAdd(&output[idx], sum / static_cast<float>(HW));
    }
}

// Host function that wraps the global average pooling kernel call.
// Assumes input tensor x has shape [N, C, H, W] and is a CUDA tensor.
at::Tensor global_avg_pool(at::Tensor x) {
    auto input = x.contiguous();
    int N = input.size(0);
    int C = input.size(1);
    int H = input.size(2);
    int W = input.size(3);
    int HW = H * W;

    // Allocate output tensor of shape [N, C]
    at::Tensor output = at::zeros({N, C}, input.options());

    // Launch kernel: one block per output element, using one warp (32 threads) per block
    int threads = 32; // warp size
    int blocks = N * C;
    globalAvgPoolKernel<<<blocks, threads>>>(input.data_ptr<float>(), output.data_ptr<float>(), C, HW);
    // cudaDeviceSynchronize();

    return output;
}

// Dense layer used in DenseNet121: batch norm, relu, convolution, and dropout
at::Tensor dense_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
    x = at::batch_norm(
        x, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::conv2d(x, conv_weight, /*bias=*/{}, /*stride=*/{1, 1}, /*padding=*/{1, 1});
    x = at::dropout(x, /*p=*/0.0, is_training);
    return x;
}

// Transition layer used between dense blocks in DenseNet121
at::Tensor transition_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
    x = at::batch_norm(
        x, bn_weight, bn_bias, bn_mean, bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::conv2d(x, conv_weight);
    x = at::avg_pool2d(x, /*kernel_size=*/{2, 2}, /*stride=*/{2, 2});
    return x;
}

// Main module function implementing DenseNet121 forward pass using the custom global average pooling kernel
at::Tensor module_fn(
    at::Tensor x,
    py::object params,
    bool is_training
) {
    // Helper lambda to retrieve parameters
    auto get_param = [&](const std::string& key) -> at::Tensor {
        return params.attr("__getitem__")(key.c_str()).cast<at::Tensor>();
    };

    // Initial features
    auto features_conv_weight = get_param("features_conv_weight");
    x = at::conv2d(x, features_conv_weight, /*bias=*/{}, /*stride=*/{2, 2}, /*padding=*/{3, 3});

    auto features_bn_mean = get_param("features_bn_mean");
    auto features_bn_var = get_param("features_bn_var");
    auto features_bn_weight = get_param("features_bn_weight");
    auto features_bn_bias = get_param("features_bn_bias");

    x = at::batch_norm(
        x, features_bn_weight, features_bn_bias, features_bn_mean, features_bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    x = at::max_pool2d(x, /*kernel_size=*/{3, 3}, /*stride=*/{2, 2}, /*padding=*/{1, 1});

    std::vector<int> num_layers = {6, 12, 24, 16};  // Layers per block for DenseNet121

    // Dense blocks and transitions
    for (int i = 0; i < 4; ++i) {
        std::vector<at::Tensor> features;
        features.push_back(x);

        for (int j = 0; j < num_layers[i]; ++j) {
            std::string prefix = "block" + std::to_string(i) + "_layer" + std::to_string(j) + "_";

            auto bn_weight = get_param(prefix + "bn_weight");
            auto bn_bias = get_param(prefix + "bn_bias");
            auto bn_mean = get_param(prefix + "bn_mean");
            auto bn_var = get_param(prefix + "bn_var");
            auto conv_weight = get_param(prefix + "conv_weight");

            at::Tensor new_feature = dense_layer_fn(
                x,
                bn_weight,
                bn_bias,
                bn_mean,
                bn_var,
                conv_weight,
                is_training
            );

            features.push_back(new_feature);

            x = at::cat(features, 1);
        }

        if (i != 3) {  // Apply transition after all blocks except the last
            std::string prefix = "transition" + std::to_string(i) + "_";

            auto bn_weight = get_param(prefix + "bn_weight");
            auto bn_bias = get_param(prefix + "bn_bias");
            auto bn_mean = get_param(prefix + "bn_mean");
            auto bn_var = get_param(prefix + "bn_var");
            auto conv_weight = get_param(prefix + "conv_weight");

            x = transition_layer_fn(
                x,
                bn_weight,
                bn_bias,
                bn_mean,
                bn_var,
                conv_weight,
                is_training
            );
        }
    }

    // Final layers
    auto final_bn_mean = get_param("final_bn_mean");
    auto final_bn_var = get_param("final_bn_var");
    auto final_bn_weight = get_param("final_bn_weight");
    auto final_bn_bias = get_param("final_bn_bias");

    x = at::batch_norm(
        x, final_bn_weight, final_bn_bias, final_bn_mean, final_bn_var,
        is_training, 0.1, 1e-5, true
    );
    x = at::relu(x);
    
    // Use custom global average pooling kernel with warp-level primitives when running on CUDA
    if (x.is_cuda()) {
        // x is of shape [N, C, H, W]. Our kernel computes [N, C].
        x = global_avg_pool(x).view({x.size(0), -1});
    } else {
        // Fallback to PyTorch's adaptive average pooling on CPU
        x = at::adaptive_avg_pool2d(x, {1, 1}).reshape({x.size(0), -1});
    }

    auto classifier_weight = get_param("classifier_weight");
    auto classifier_bias = get_param("classifier_bias");
    x = at::linear(x, classifier_weight, classifier_bias);

    return x;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn, "DenseNet121 forward with warp-level reduction in avg pooling");
}
